# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial version.

Revision ID: 27e647c0fad4
Revises:
Create Date: 2021-12-23 11:13:26.305412
"""

import textwrap

from alembic import op
from oslo_log import log
from oslo_utils import timeutils
import sqlalchemy as sql

from keystone.assignment.backends import sql as assignment_sql
from keystone.common import sql as ks_sql
import keystone.conf
from keystone.identity.mapping_backends import mapping as mapping_backend

# revision identifiers, used by Alembic.
revision = '27e647c0fad4'
down_revision = None
depends_on = None

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)

NULL_DOMAIN_ID = '<<keystone.domain.root>>'


def upgrade():
    bind = op.get_bind()

    if bind.engine.name == 'mysql':
        # Set default DB charset to UTF8.
        op.execute(
            f'ALTER DATABASE {bind.engine.url.database} DEFAULT CHARACTER SET utf8'
        )

    op.create_table(
        'application_credential',
        sql.Column(
            'internal_id', sql.Integer, primary_key=True, nullable=False
        ),
        sql.Column('id', sql.String(length=64), nullable=False),
        sql.Column('name', sql.String(length=255), nullable=False),
        sql.Column('secret_hash', sql.String(length=255), nullable=False),
        sql.Column('description', sql.Text),
        sql.Column('user_id', sql.String(length=64), nullable=False),
        sql.Column('project_id', sql.String(64), nullable=True),
        sql.Column('expires_at', ks_sql.DateTimeInt()),
        sql.Column('system', sql.String(64), nullable=True),
        sql.Column('unrestricted', sql.Boolean),
        sql.UniqueConstraint(
            'user_id', 'name', name='duplicate_app_cred_constraint'
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'assignment',
        sql.Column(
            'type',
            sql.Enum(
                assignment_sql.AssignmentType.USER_PROJECT,
                assignment_sql.AssignmentType.GROUP_PROJECT,
                assignment_sql.AssignmentType.USER_DOMAIN,
                assignment_sql.AssignmentType.GROUP_DOMAIN,
                name='type',
            ),
            nullable=False,
        ),
        sql.Column('actor_id', sql.String(64), nullable=False),
        sql.Column('target_id', sql.String(64), nullable=False),
        sql.Column('role_id', sql.String(64), nullable=False),
        sql.Column('inherited', sql.Boolean, default=False, nullable=False),
        sql.PrimaryKeyConstraint(
            'type', 'actor_id', 'target_id', 'role_id', 'inherited'
        ),
        sql.Index('ix_actor_id', 'actor_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'access_rule',
        sql.Column('id', sql.Integer, primary_key=True, nullable=False),
        sql.Column('service', sql.String(64)),
        sql.Column('path', sql.String(128)),
        sql.Column('method', sql.String(16)),
        sql.Column('external_id', sql.String(64)),
        sql.Column('user_id', sql.String(64)),
        sql.UniqueConstraint(
            'external_id', name='access_rule_external_id_key'
        ),
        sql.UniqueConstraint(
            'user_id',
            'service',
            'path',
            'method',
            name='duplicate_access_rule_for_user_constraint',
        ),
        sql.Index('user_id', 'user_id'),
        sql.Index('external_id', 'external_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'config_register',
        sql.Column('type', sql.String(64), primary_key=True),
        sql.Column('domain_id', sql.String(64), nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'consumer',
        sql.Column('id', sql.String(64), primary_key=True, nullable=False),
        sql.Column('description', sql.String(64), nullable=True),
        sql.Column('secret', sql.String(64), nullable=False),
        sql.Column('extra', sql.Text(), nullable=False),
    )

    op.create_table(
        'credential',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('user_id', sql.String(length=64), nullable=False),
        sql.Column('project_id', sql.String(length=64)),
        sql.Column('type', sql.String(length=255), nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('key_hash', sql.String(64), nullable=False),
        sql.Column('encrypted_blob', ks_sql.Text, nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'group',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('domain_id', sql.String(length=64), nullable=False),
        sql.Column('name', sql.String(length=64), nullable=False),
        sql.Column('description', sql.Text),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.UniqueConstraint(
            'domain_id', 'name', name='ixu_group_name_domain_id'
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'id_mapping',
        sql.Column('public_id', sql.String(64), primary_key=True),
        sql.Column('domain_id', sql.String(64), nullable=False),
        sql.Column('local_id', sql.String(255), nullable=False),
        sql.Column(
            'entity_type',
            sql.Enum(
                mapping_backend.EntityType.USER,
                mapping_backend.EntityType.GROUP,
                name='entity_type',
            ),
            nullable=False,
        ),
        sql.UniqueConstraint(
            'domain_id', 'local_id', 'entity_type', name='domain_id'
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'identity_provider',
        sql.Column('id', sql.String(64), primary_key=True),
        sql.Column('enabled', sql.Boolean, nullable=False),
        sql.Column('description', sql.Text(), nullable=True),
        sql.Column('domain_id', sql.String(64), nullable=False),
        sql.Column('authorization_ttl', sql.Integer, nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'idp_remote_ids',
        sql.Column(
            'idp_id',
            sql.String(64),
            sql.ForeignKey('identity_provider.id', ondelete='CASCADE'),
        ),
        sql.Column('remote_id', sql.String(255), primary_key=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'mapping',
        sql.Column('id', sql.String(64), primary_key=True),
        sql.Column('rules', sql.Text(), nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'policy',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('type', sql.String(length=255), nullable=False),
        sql.Column('blob', ks_sql.JsonBlob, nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'policy_association',
        sql.Column('id', sql.String(64), primary_key=True),
        sql.Column('policy_id', sql.String(64), nullable=False),
        sql.Column('endpoint_id', sql.String(64), nullable=True),
        sql.Column('service_id', sql.String(64), nullable=True),
        sql.Column('region_id', sql.String(64), nullable=True),
        sql.UniqueConstraint('endpoint_id', 'service_id', 'region_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'project',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('name', sql.String(length=64), nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('description', sql.Text),
        sql.Column('enabled', sql.Boolean),
        sql.Column(
            'domain_id',
            sql.String(length=64),
            sql.ForeignKey('project.id', name='project_domain_id_fkey'),
            nullable=False,
        ),
        sql.Column(
            'parent_id',
            sql.String(64),
            sql.ForeignKey('project.id', name='project_parent_id_fkey'),
            nullable=True,
        ),
        sql.Column(
            'is_domain',
            sql.Boolean,
            nullable=False,
            server_default='0',
            default=False,
        ),
        sql.UniqueConstraint(
            'domain_id', 'name', name='ixu_project_name_domain_id'
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'project_endpoint',
        sql.Column(
            'endpoint_id', sql.String(64), primary_key=True, nullable=False
        ),
        sql.Column(
            'project_id', sql.String(64), primary_key=True, nullable=False
        ),
    )

    op.create_table(
        'project_option',
        sql.Column(
            'project_id',
            sql.String(64),
            sql.ForeignKey('project.id', ondelete='CASCADE'),
            nullable=False,
            primary_key=True,
        ),
        sql.Column(
            'option_id', sql.String(4), nullable=False, primary_key=True
        ),
        sql.Column('option_value', ks_sql.JsonBlob, nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    # NOTE(lamt) To allow tag name to be case sensitive for MySQL, the 'name'
    # column needs to use collation, which is incompatible with Postgresql.
    # Using unicode to mirror nova's server tag:
    # https://github.com/openstack/nova/blob/master/nova/db/sqlalchemy/models.py
    op.create_table(
        'project_tag',
        sql.Column(
            'project_id',
            sql.String(64),
            sql.ForeignKey('project.id', ondelete='CASCADE'),
            nullable=False,
            primary_key=True,
        ),
        sql.Column('name', sql.Unicode(255), nullable=False, primary_key=True),
        sql.UniqueConstraint('project_id', 'name'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'region',
        sql.Column('id', sql.String(255), primary_key=True),
        sql.Column('description', sql.String(255), nullable=False),
        sql.Column('parent_region_id', sql.String(255), nullable=True),
        sql.Column('extra', sql.Text()),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'registered_limit',
        sql.Column('id', sql.String(length=64), nullable=False),
        sql.Column('service_id', sql.String(255)),
        sql.Column('region_id', sql.String(64), nullable=True),
        sql.Column('resource_name', sql.String(255)),
        sql.Column('default_limit', sql.Integer, nullable=False),
        sql.Column('description', sql.Text),
        sql.Column('internal_id', sql.Integer, primary_key=True),
        # NOTE(stephenfin): Name chosen to preserve backwards compatibility
        # with names used for primary key unique constraints
        sql.UniqueConstraint('id', name='registered_limit_id_key'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'request_token',
        sql.Column('id', sql.String(64), primary_key=True, nullable=False),
        sql.Column('request_secret', sql.String(64), nullable=False),
        sql.Column('verifier', sql.String(64), nullable=True),
        sql.Column('authorizing_user_id', sql.String(64), nullable=True),
        sql.Column('requested_project_id', sql.String(64), nullable=False),
        sql.Column('role_ids', sql.Text(), nullable=True),
        sql.Column(
            'consumer_id',
            sql.String(64),
            sql.ForeignKey('consumer.id'),
            nullable=False,
            index=True,
        ),
        sql.Column('expires_at', sql.String(64), nullable=True),
    )

    op.create_table(
        'revocation_event',
        sql.Column('id', sql.Integer, primary_key=True),
        sql.Column('domain_id', sql.String(64)),
        sql.Column('project_id', sql.String(64)),
        sql.Column('user_id', sql.String(64)),
        sql.Column('role_id', sql.String(64)),
        sql.Column('trust_id', sql.String(64)),
        sql.Column('consumer_id', sql.String(64)),
        sql.Column('access_token_id', sql.String(64)),
        sql.Column('issued_before', sql.DateTime(), nullable=False),
        sql.Column('expires_at', sql.DateTime()),
        sql.Column('revoked_at', sql.DateTime(), nullable=False),
        sql.Column('audit_id', sql.String(32), nullable=True),
        sql.Column('audit_chain_id', sql.String(32), nullable=True),
        # NOTE(stephenfin): The '_new' suffix here is due to migration 095,
        # which changed the 'id' column from String(64) to Integer. It did this
        # by creating a 'revocation_event_new' table and populating it with
        # data from the 'revocation_event' table before deleting the
        # 'revocation_event' table and renaming the 'revocation_event_new'
        # table to 'revocation_event'. Because the 'revoked_at' column had
        # 'index=True', sqlalchemy automatically generated the index name as
        # 'ix_{table}_{column}'. However, when intitially created, '{table}'
        # was 'revocation_event_new' so the index got that name. We may wish to
        # rename this eventually.
        sql.Index('ix_revocation_event_new_revoked_at', 'revoked_at'),
        sql.Index('ix_revocation_event_issued_before', 'issued_before'),
        sql.Index(
            'ix_revocation_event_project_id_issued_before',
            'project_id',
            'issued_before',
        ),
        sql.Index(
            'ix_revocation_event_user_id_issued_before',
            'user_id',
            'issued_before',
        ),
        sql.Index(
            'ix_revocation_event_audit_id_issued_before',
            'audit_id',
            'issued_before',
        ),
    )

    op.create_table(
        'role',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('name', sql.String(length=255), nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column(
            'domain_id',
            sql.String(64),
            nullable=False,
            server_default='<<null>>',
        ),
        sql.Column('description', sql.String(255), nullable=True),
        sql.UniqueConstraint(
            'name', 'domain_id', name='ixu_role_name_domain_id'
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'role_option',
        sql.Column(
            'role_id',
            sql.String(64),
            sql.ForeignKey('role.id', ondelete='CASCADE'),
            nullable=False,
            primary_key=True,
        ),
        sql.Column(
            'option_id', sql.String(4), nullable=False, primary_key=True
        ),
        sql.Column('option_value', ks_sql.JsonBlob, nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'sensitive_config',
        sql.Column('domain_id', sql.String(64), primary_key=True),
        sql.Column('group', sql.String(255), primary_key=True),
        sql.Column('option', sql.String(255), primary_key=True),
        sql.Column('value', ks_sql.JsonBlob.impl, nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'service',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('type', sql.String(length=255)),
        sql.Column(
            'enabled',
            sql.Boolean,
            nullable=False,
            default=True,
            server_default='1',
        ),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'service_provider',
        sql.Column('auth_url', sql.String(256), nullable=False),
        sql.Column('id', sql.String(64), primary_key=True),
        sql.Column('enabled', sql.Boolean, nullable=False),
        sql.Column('description', sql.Text(), nullable=True),
        sql.Column('sp_url', sql.String(256), nullable=False),
        sql.Column(
            'relay_state_prefix',
            sql.String(256),
            nullable=False,
            server_default=CONF.saml.relay_state_prefix,
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'system_assignment',
        sql.Column('type', sql.String(64), nullable=False),
        sql.Column('actor_id', sql.String(64), nullable=False),
        sql.Column('target_id', sql.String(64), nullable=False),
        sql.Column('role_id', sql.String(64), nullable=False),
        sql.Column('inherited', sql.Boolean, default=False, nullable=False),
        sql.PrimaryKeyConstraint(
            'type', 'actor_id', 'target_id', 'role_id', 'inherited'
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'token',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('expires', sql.DateTime, default=None),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('valid', sql.Boolean, default=True, nullable=False),
        sql.Column('trust_id', sql.String(length=64)),
        sql.Column('user_id', sql.String(length=64)),
        sql.Index('ix_token_expires', 'expires'),
        sql.Index('ix_token_expires_valid', 'expires', 'valid'),
        sql.Index('ix_token_user_id', 'user_id'),
        sql.Index('ix_token_trust_id', 'trust_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'trust',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('trustor_user_id', sql.String(length=64), nullable=False),
        sql.Column('trustee_user_id', sql.String(length=64), nullable=False),
        sql.Column('project_id', sql.String(length=64)),
        sql.Column('impersonation', sql.Boolean, nullable=False),
        sql.Column('deleted_at', sql.DateTime),
        sql.Column('expires_at', sql.DateTime),
        sql.Column('remaining_uses', sql.Integer, nullable=True),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('expires_at_int', ks_sql.DateTimeInt()),
        sql.UniqueConstraint(
            'trustor_user_id',
            'trustee_user_id',
            'project_id',
            'impersonation',
            'expires_at',
            'expires_at_int',
            name='duplicate_trust_constraint_expanded',
        ),
        sql.Column('redelegated_trust_id', sql.String(64), nullable=True),
        sql.Column('redelegation_count', sql.Integer, nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'trust_role',
        sql.Column(
            'trust_id', sql.String(length=64), primary_key=True, nullable=False
        ),
        sql.Column(
            'role_id', sql.String(length=64), primary_key=True, nullable=False
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'user',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('enabled', sql.Boolean),
        sql.Column('default_project_id', sql.String(length=64)),
        sql.Column('created_at', sql.DateTime(), nullable=True),
        sql.Column('last_active_at', sql.Date(), nullable=True),
        sql.Column('domain_id', sql.String(64), nullable=False),
        sql.UniqueConstraint('id', 'domain_id', name='ixu_user_id_domain_id'),
        sql.Index('ix_default_project_id', 'default_project_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'user_group_membership',
        sql.Column(
            'user_id',
            sql.String(length=64),
            sql.ForeignKey('user.id', name='fk_user_group_membership_user_id'),
            primary_key=True,
        ),
        sql.Column(
            'group_id',
            sql.String(length=64),
            sql.ForeignKey(
                'group.id', name='fk_user_group_membership_group_id'
            ),
            primary_key=True,
        ),
        # NOTE(stevemar): The index was named 'group_id' in
        # 050_fk_consistent_indexes.py and needs to be preserved
        sql.Index('group_id', 'group_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'user_option',
        sql.Column(
            'user_id',
            sql.String(64),
            sql.ForeignKey('user.id', ondelete='CASCADE'),
            nullable=False,
            primary_key=True,
        ),
        sql.Column(
            'option_id', sql.String(4), nullable=False, primary_key=True
        ),
        sql.Column('option_value', ks_sql.JsonBlob, nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'whitelisted_config',
        sql.Column('domain_id', sql.String(64), primary_key=True),
        sql.Column('group', sql.String(255), primary_key=True),
        sql.Column('option', sql.String(255), primary_key=True),
        sql.Column('value', ks_sql.JsonBlob.impl, nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'access_token',
        sql.Column('id', sql.String(64), primary_key=True, nullable=False),
        sql.Column('access_secret', sql.String(64), nullable=False),
        sql.Column(
            'authorizing_user_id', sql.String(64), nullable=False, index=True
        ),
        sql.Column('project_id', sql.String(64), nullable=False),
        sql.Column('role_ids', sql.Text(), nullable=False),
        sql.Column(
            'consumer_id',
            sql.String(64),
            sql.ForeignKey('consumer.id'),
            nullable=False,
            index=True,
        ),
        sql.Column('expires_at', sql.String(64), nullable=True),
    )

    op.create_table(
        'application_credential_role',
        sql.Column(
            'application_credential_id',
            sql.Integer,
            sql.ForeignKey(
                'application_credential.internal_id', ondelete='CASCADE'
            ),
            primary_key=True,
            nullable=False,
        ),
        sql.Column(
            'role_id', sql.String(length=64), primary_key=True, nullable=False
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'application_credential_access_rule',
        sql.Column(
            'application_credential_id',
            sql.Integer,
            sql.ForeignKey(
                'application_credential.internal_id', ondelete='CASCADE'
            ),
            primary_key=True,
            nullable=False,
        ),
        sql.Column(
            'access_rule_id',
            sql.Integer,
            sql.ForeignKey('access_rule.id', ondelete='CASCADE'),
            primary_key=True,
            nullable=False,
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'endpoint',
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('legacy_endpoint_id', sql.String(length=64)),
        sql.Column('interface', sql.String(length=8), nullable=False),
        sql.Column(
            'service_id',
            sql.String(length=64),
            sql.ForeignKey('service.id', name='endpoint_service_id_fkey'),
            nullable=False,
        ),
        sql.Column('url', sql.Text, nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column(
            'enabled',
            sql.Boolean,
            nullable=False,
            default=True,
            server_default='1',
        ),
        sql.Column(
            'region_id',
            sql.String(length=255),
            sql.ForeignKey('region.id', name='fk_endpoint_region_id'),
            nullable=True,
        ),
        # NOTE(stevemar): The index was named 'service_id' in
        # 050_fk_consistent_indexes.py and needs to be preserved
        sql.Index('service_id', 'service_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'endpoint_group',
        sql.Column('id', sql.String(64), primary_key=True),
        sql.Column('name', sql.String(255), nullable=False),
        sql.Column('description', sql.Text, nullable=True),
        sql.Column('filters', sql.Text(), nullable=False),
    )

    op.create_table(
        'expiring_user_group_membership',
        sql.Column(
            'user_id',
            sql.String(64),
            sql.ForeignKey('user.id'),
            primary_key=True,
        ),
        sql.Column(
            'group_id',
            sql.String(64),
            sql.ForeignKey('group.id'),
            primary_key=True,
        ),
        sql.Column(
            'idp_id',
            sql.String(64),
            sql.ForeignKey('identity_provider.id', ondelete='CASCADE'),
            primary_key=True,
        ),
        sql.Column('last_verified', sql.DateTime(), nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'federation_protocol',
        sql.Column('id', sql.String(64), primary_key=True),
        sql.Column(
            'idp_id',
            sql.String(64),
            sql.ForeignKey('identity_provider.id', ondelete='CASCADE'),
            primary_key=True,
        ),
        sql.Column('mapping_id', sql.String(64), nullable=False),
        sql.Column('remote_id_attribute', sql.String(64)),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'implied_role',
        sql.Column(
            'prior_role_id',
            sql.String(length=64),
            sql.ForeignKey(
                'role.id',
                name='implied_role_prior_role_id_fkey',
                ondelete='CASCADE',
            ),
            primary_key=True,
        ),
        sql.Column(
            'implied_role_id',
            sql.String(length=64),
            sql.ForeignKey(
                'role.id',
                name='implied_role_implied_role_id_fkey',
                ondelete='CASCADE',
            ),
            primary_key=True,
        ),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'limit',
        sql.Column('id', sql.String(length=64), nullable=False),
        sql.Column('project_id', sql.String(64), nullable=True),
        sql.Column('resource_limit', sql.Integer, nullable=False),
        sql.Column('description', sql.Text),
        sql.Column('internal_id', sql.Integer, primary_key=True),
        # FIXME(stephenfin): This should have a foreign key constraint on
        # registered_limit.id, but sqlalchemy-migrate clearly didn't handle
        # creating a column with embedded FK info as was attempted in 048
        sql.Column('registered_limit_id', sql.String(64)),
        sql.Column('domain_id', sql.String(64), nullable=True),
        # NOTE(stephenfin): Name chosen to preserve backwards compatibility
        # with names used for primary key unique constraints
        sql.UniqueConstraint('id', name='limit_id_key'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'local_user',
        sql.Column('id', sql.Integer, primary_key=True, nullable=False),
        sql.Column('user_id', sql.String(64), nullable=False, unique=True),
        sql.Column('domain_id', sql.String(64), nullable=False),
        sql.Column('name', sql.String(255), nullable=False),
        sql.Column('failed_auth_count', sql.Integer, nullable=True),
        sql.Column('failed_auth_at', sql.DateTime(), nullable=True),
        sql.ForeignKeyConstraint(
            ['user_id', 'domain_id'],
            ['user.id', 'user.domain_id'],
            name='local_user_user_id_fkey',
            onupdate='CASCADE',
            ondelete='CASCADE',
        ),
        sql.UniqueConstraint('domain_id', 'name'),
    )

    op.create_table(
        'nonlocal_user',
        sql.Column('domain_id', sql.String(64), primary_key=True),
        sql.Column('name', sql.String(255), primary_key=True),
        sql.Column('user_id', sql.String(64), nullable=False),
        sql.ForeignKeyConstraint(
            ['user_id', 'domain_id'],
            ['user.id', 'user.domain_id'],
            name='nonlocal_user_user_id_fkey',
            onupdate='CASCADE',
            ondelete='CASCADE',
        ),
        sql.UniqueConstraint('user_id', name='ixu_nonlocal_user_user_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    op.create_table(
        'password',
        sql.Column('id', sql.Integer, primary_key=True, nullable=False),
        sql.Column(
            'local_user_id',
            sql.Integer,
            sql.ForeignKey('local_user.id', ondelete='CASCADE'),
            nullable=False,
        ),
        sql.Column('expires_at', sql.DateTime(), nullable=True),
        sql.Column(
            'self_service',
            sql.Boolean,
            nullable=False,
            server_default='0',
            default=False,
        ),
        # NOTE(notmorgan): To support the full range of scrypt and pbkfd
        # password hash lengths, this should be closer to varchar(1500) instead
        # of varchar(255).
        sql.Column('password_hash', sql.String(255), nullable=True),
        sql.Column(
            'created_at_int',
            ks_sql.DateTimeInt(),
            nullable=False,
            default=0,
            server_default='0',
        ),
        sql.Column('expires_at_int', ks_sql.DateTimeInt(), nullable=True),
        sql.Column(
            'created_at',
            sql.DateTime(),
            nullable=False,
            default=timeutils.utcnow,
        ),
    )

    op.create_table(
        'project_endpoint_group',
        sql.Column(
            'endpoint_group_id',
            sql.String(64),
            sql.ForeignKey('endpoint_group.id'),
            nullable=False,
        ),
        sql.Column('project_id', sql.String(64), nullable=False),
        sql.PrimaryKeyConstraint('endpoint_group_id', 'project_id'),
    )

    op.create_table(
        'federated_user',
        sql.Column('id', sql.Integer, primary_key=True, nullable=False),
        sql.Column(
            'user_id',
            sql.String(64),
            sql.ForeignKey('user.id', ondelete='CASCADE'),
            nullable=False,
        ),
        sql.Column(
            'idp_id',
            sql.String(64),
            sql.ForeignKey('identity_provider.id', ondelete='CASCADE'),
            nullable=False,
        ),
        sql.Column('protocol_id', sql.String(64), nullable=False),
        sql.Column('unique_id', sql.String(255), nullable=False),
        sql.Column('display_name', sql.String(255), nullable=True),
        sql.ForeignKeyConstraint(
            ['protocol_id', 'idp_id'],
            ['federation_protocol.id', 'federation_protocol.idp_id'],
            name='federated_user_protocol_id_fkey',
            ondelete='CASCADE',
        ),
        sql.UniqueConstraint('idp_id', 'protocol_id', 'unique_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )

    if bind.engine.name == 'sqlite':
        # NOTE(stevemar): We need to keep this FK constraint due to 073, but
        # only for sqlite, once we collapse 073 we can remove this constraint
        with op.batch_alter_table('assignment') as batch_op:
            batch_op.create_foreign_key(
                'fk_assignment_role_id', 'role', ['role_id'], ['id']
            )

    # TODO(stephenfin): Remove these procedures in a future contract migration

    if bind.engine.name == 'postgresql':
        error_message = (
            'Credential migration in progress. Cannot perform '
            'writes to credential table.'
        )
        credential_update_trigger = textwrap.dedent(
            f"""
        CREATE OR REPLACE FUNCTION keystone_read_only_update()
          RETURNS trigger AS
        $BODY$
        BEGIN
          IF NEW.encrypted_blob IS NULL THEN
            RAISE EXCEPTION '{error_message}';
          END IF;
          IF NEW.encrypted_blob IS NOT NULL AND OLD.blob IS NULL THEN
            RAISE EXCEPTION '{error_message}';
          END IF;
          RETURN NEW;
        END
        $BODY$ LANGUAGE plpgsql;
        """
        )
        op.execute(credential_update_trigger)

        error_message = (
            'Identity provider migration in progress. Cannot '
            'insert new rows into the identity_provider table at '
            'this time.'
        )
        identity_provider_insert_trigger = textwrap.dedent(
            f"""
        CREATE OR REPLACE FUNCTION keystone_read_only_insert()
          RETURNS trigger AS
        $BODY$
        BEGIN
          RAISE EXCEPTION '{error_message}';
        END
        $BODY$ LANGUAGE plpgsql;
        """
        )
        op.execute(identity_provider_insert_trigger)

        federated_user_insert_trigger = textwrap.dedent(
            """
        CREATE OR REPLACE FUNCTION update_federated_user_domain_id()
            RETURNS trigger AS
        $BODY$
        BEGIN
            UPDATE "user" SET domain_id = (
                SELECT domain_id FROM identity_provider WHERE id = NEW.idp_id)
                WHERE id = NEW.user_id and domain_id IS NULL;
            RETURN NULL;
        END
        $BODY$ LANGUAGE plpgsql;
        """
        )
        op.execute(federated_user_insert_trigger)

        local_user_insert_trigger = textwrap.dedent(
            """
        CREATE OR REPLACE FUNCTION update_user_domain_id()
            RETURNS trigger AS
        $BODY$
        BEGIN
            UPDATE "user" SET domain_id = NEW.domain_id
                WHERE id = NEW.user_id;
            RETURN NULL;
        END
        $BODY$ LANGUAGE plpgsql;
        """
        )
        op.execute(local_user_insert_trigger)

    # FIXME(stephenfin): Remove these indexes. They're left over from attempts
    # to remove foreign key constraints in past migrations. Apparently
    # sqlalchemy-migrate didn't do the job fully and left behind indexes
    if bind.engine.name == 'mysql':
        op.create_index('region_id', 'registered_limit', ['region_id'])

        # FIXME(stephenfin): This should be dropped when we add the FK
        # constraint to this column
        op.create_index(
            'registered_limit_id', 'limit', ['registered_limit_id']
        )

        # FIXME(stephenfin): These are leftover from when we removed a FK
        # constraint and should probable be dropped
        op.create_index('domain_id', 'identity_provider', ['domain_id'])
        op.create_index('domain_id', 'user', ['domain_id'])

    # data migration

    def _generate_root_domain_project():
        # Generate a project that will act as a root for all domains, in order
        # for use to be able to use a FK constraint on domain_id. Projects
        # acting as a domain will not reference this as their parent_id, just
        # as domain_id.
        #
        # This special project is filtered out by the driver, so is never
        # visible to the manager or API.

        project_ref = {
            'id': NULL_DOMAIN_ID,
            'name': NULL_DOMAIN_ID,
            'enabled': False,
            'description': '',
            'domain_id': NULL_DOMAIN_ID,
            'is_domain': True,
            'parent_id': None,
            'extra': '{}',
        }
        return project_ref

    bind = op.get_bind()
    meta = sql.MetaData()
    project = sql.Table('project', meta, autoload_with=bind)

    root_domain_project = _generate_root_domain_project()
    op.execute(project.insert().values(**root_domain_project))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Options specific to resources managed by Keystone (Domain, User, etc)."""

from keystone.common import validation
from keystone.i18n import _


def _validator(value):
    return


def boolean_validator(value):
    if value not in (True, False):
        raise TypeError(_('Expected boolean value, got %r') % type(value))


def ref_mapper_to_dict_options(ref):
    """Convert the values in _resource_option_mapper to options dict.

    NOTE: this is to be called from the relevant `to_dict` methods or
          similar and must be called from within the active session context.

    :param ref: the DB model ref to extract options from
    :returns: Dict of options as expected to be returned out of to_dict in
              the `options` key.
    """
    options = {}
    for opt in ref._resource_option_mapper.values():
        if opt.option_id in ref.resource_options_registry.option_ids:
            r_opt = ref.resource_options_registry.get_option_by_id(
                opt.option_id
            )
            if r_opt is not None:
                options[r_opt.option_name] = opt.option_value
    return options


def get_resource_option(model, option_id):
    """Get the resource option information from the model's mapper."""
    if option_id in model._resource_option_mapper.keys():
        return model._resource_option_mapper[option_id]
    return None


def resource_options_ref_to_mapper(ref, option_class):
    """Convert the _resource_options property-dict to options attr map.

    The model must have the resource option mapper located in the
    ``_resource_option_mapper`` attribute.

    The model must have the resource option registry located in the
    ``resource_options_registry`` attribute.

    The option dict with key(opt_id), value(opt_value) will be pulled from
    ``ref._resource_options``.

    NOTE: This function MUST be called within the active writer session
          context!

    :param ref: The DB model reference that is actually stored to the
                backend.
    :param option_class: Class that is used to store the resource option
                         in the DB.
    """
    options = getattr(ref, '_resource_options', None)
    if options is not None:
        # To ensure everything is clean, no lingering refs.
        delattr(ref, '_resource_options')
    else:
        # _resource_options didn't exist. Work from an empty set.
        options = {}

    # NOTE(notmorgan): explicitly use .keys() here as the attribute mapper
    # has some oddities at times. This guarantees we are working with keys.
    set_options = set(ref._resource_option_mapper.keys())
    # Get any options that are not registered and slate them for removal from
    # the DB. This will delete unregistered options.
    clear_options = set_options.difference(
        ref.resource_options_registry.option_ids
    )
    options.update({x: None for x in clear_options})

    # Set the resource options for user in the Attribute Mapping.
    for r_opt_id, r_opt_value in options.items():
        if r_opt_value is None:
            # Delete any option set explicitly to None, ignore unset
            # options.
            ref._resource_option_mapper.pop(r_opt_id, None)
        else:
            # Set any options on the user_ref itself.
            opt_obj = option_class(
                option_id=r_opt_id, option_value=r_opt_value
            )
            ref._resource_option_mapper[r_opt_id] = opt_obj


class ResourceOptionRegistry:
    def __init__(self, registry_name):
        self._registered_options = {}
        self._registry_type = registry_name

    @property
    def option_names(self):
        return {opt.option_name for opt in self.options}

    @property
    def options_by_name(self):
        return {
            opt.option_name: opt for opt in self._registered_options.values()
        }

    @property
    def options(self):
        return self._registered_options.values()

    @property
    def option_ids(self):
        return set(self._registered_options.keys())

    def get_option_by_id(self, opt_id):
        return self._registered_options.get(opt_id, None)

    def get_option_by_name(self, name):
        for option in self._registered_options.values():
            if name == option.option_name:
                return option
        return None

    @property
    def json_schema(self):
        schema = {
            'type': 'object',
            'properties': {},
            'additionalProperties': False,
        }
        for opt in self.options:
            if opt.json_schema is not None:
                # NOTE(notmorgan): All options are nullable. Null indicates
                # the option should be reset and removed from the DB store.
                schema['properties'][opt.option_name] = validation.nullable(
                    opt.json_schema
                )
            else:
                # NOTE(notmorgan): without 'type' being specified, this
                # can be of any-type. We are simply specifying no interesting
                # values beyond that the property may exist here.
                schema['properties'][opt.option_name] = {}
        return schema

    def register_option(self, option):
        if option in self.options:
            # Re-registering the exact same option does nothing.
            return

        if option.option_id in self._registered_options:
            raise ValueError(
                _('Option %(option_id)s already defined in %(registry)s.')
                % {
                    'option_id': option.option_id,
                    'registry': self._registry_type,
                }
            )
        if option.option_name in self.option_names:
            raise ValueError(
                _('Option %(option_name)s already defined in %(registry)s')
                % {
                    'option_name': option.option_name,
                    'registry': self._registry_type,
                }
            )
        self._registered_options[option.option_id] = option


class ResourceOption:
    def __init__(
        self,
        option_id,
        option_name,
        validator=_validator,
        json_schema_validation=None,
    ):
        """The base object to define the option(s) to be stored in the DB.

        :param option_id: The ID of the option. This will be used to lookup
                          the option value from the DB and should not be
                          changed once defined as the values will no longer
                          be correctly mapped to the keys in the user_ref when
                          retrieving the data from the DB.
        :type option_id: str
        :param option_name: The name of the option. This value will be used
                            to map the value from the user request on a
                            resource update to the correct option id to be
                            stored in the database. This value should not be
                            changed once defined as it will change the
                            resulting keys in the user_ref.
        :type option_name: str
        :param validator: A callable that raises TypeError if the value to be
                          persisted is incorrect. A single argument of the
                          value to be persisted will be passed to it. No return
                          value is expected.
        :type validator: callable
        :param json_schema_validation: Dictionary defining the JSON schema
                                       validation for the option itself. This
                                       is used to generate the JSON Schema
                                       validator(s) used at the API layer
        :type json_schema_validation: dict
        """
        if not isinstance(option_id, str) and len(option_id) == 4:
            raise TypeError(
                _('`option_id` must be a string, got %r') % option_id
            )
        elif len(option_id) != 4:
            raise ValueError(
                _('`option_id` must be 4 characters in length. Got %r')
                % option_id
            )
        if not isinstance(option_name, str):
            raise TypeError(
                _('`option_name` must be a string. Got %r') % option_name
            )

        self._option_id = option_id
        self._option_name = option_name
        self.validator = validator
        self._json_schema_validation = json_schema_validation

    @property
    def json_schema(self):
        return self._json_schema_validation or None

    @property
    def option_name(self):
        # NOTE(notmorgan) Option IDs should never be set outside of definition
        # time.
        return self._option_name

    @property
    def option_id(self):
        # NOTE(notmorgan) Option IDs should never be set outside of definition
        # time.
        return self._option_id



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from oslo_serialization import jsonutils

from keystone.common.policies import user as up
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _CommonUserTests:
    """Common default functionality for all users."""

    def test_user_can_get_their_own_user_reference(self):
        with self.test_client() as c:
            r = c.get(f'/v3/users/{self.user_id}', headers=self.headers)
            self.assertEqual(self.user_id, r.json['user']['id'])


class _SystemUserTests:
    """Common default functionality for all system users."""

    def test_user_can_get_other_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            r = c.get('/v3/users/{}'.format(user['id']), headers=self.headers)
            self.assertEqual(user['id'], r.json['user']['id'])

    def test_user_cannot_get_non_existent_user_not_found(self):
        with self.test_client() as c:
            c.get(
                f'/v3/users/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND,
            )

    def test_user_can_list_users(self):
        expected_user_ids = []
        for _ in range(3):
            user = PROVIDERS.identity_api.create_user(
                unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
            )
            expected_user_ids.append(user['id'])

        with self.test_client() as c:
            r = c.get('/v3/users', headers=self.headers)
            returned_user_ids = []
            for user in r.json['users']:
                returned_user_ids.append(user['id'])

            for user_id in expected_user_ids:
                self.assertIn(user_id, returned_user_ids)


class _SystemMemberAndReaderUserTests:
    """Common functionality for system readers and system members."""

    def test_user_cannot_create_users(self):
        create = {
            'user': {
                'name': uuid.uuid4().hex,
                'domain': CONF.identity.default_domain_id,
            }
        }

        with self.test_client() as c:
            c.post(
                '/v3/users',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            update = {'user': {'email': uuid.uuid4().hex}}

            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_non_existent_user_forbidden(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_non_existent_user_forbidden(self):
        with self.test_client() as c:
            c.delete(
                f'/v3/users/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _DomainUserTests:
    """Commont default functionality for all domain users."""

    def test_user_can_get_user_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            r = c.get('/v3/users/{}'.format(user['id']), headers=self.headers)
            self.assertEqual(user['id'], r.json['user']['id'])

    def test_user_cannot_get_user_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.get(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_list_users_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            r = c.get('/v3/users', headers=self.headers)
            self.assertEqual(2, len(r.json['users']))
            user_ids = []
            for user in r.json['users']:
                user_ids.append(user['id'])
            self.assertIn(self.user_id, user_ids)
            self.assertIn(user['id'], user_ids)

    def test_user_cannot_list_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            r = c.get('/v3/users', headers=self.headers)
            user_ids = []
            for u in r.json['users']:
                user_ids.append(u['id'])
            self.assertNotIn(user['id'], user_ids)


class _DomainMemberAndReaderUserTests:
    """Functionality for all domain members and domain readers."""

    def test_user_cannot_create_users_within_domain(self):
        create = {
            'user': {'domain_id': self.domain_id, 'name': uuid.uuid4().hex}
        }

        with self.test_client() as c:
            c.post(
                '/v3/users',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_create_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        create = {
            'user': {'domain_id': domain['id'], 'name': uuid.uuid4().hex}
        }

        with self.test_client() as c:
            c.post(
                '/v3/users',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_users_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_non_existent_user_forbidden(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_users_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_non_existent_user_forbidden(self):
        with self.test_client() as c:
            c.delete(
                f'/v3/users/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _ProjectUserTests:
    """Common tests cases for all project users."""

    def test_user_cannot_get_users_within_their_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.get(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_users_in_other_domains(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.get(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_non_existent_user_forbidden(self):
        with self.test_client() as c:
            c.get(
                f'/v3/users/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_users_within_domain(self):
        with self.test_client() as c:
            c.get(
                f'/v3/users?domain_id={self.domain_id}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_users_in_other_domains(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.get(
                '/v3/users?domain_id={}'.format(domain['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_create_users_within_domain(self):
        create = {
            'user': {'domain_id': self.domain_id, 'name': uuid.uuid4().hex}
        }

        with self.test_client() as c:
            c.post(
                '/v3/users',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_create_users_in_other_domains(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        create = {
            'user': {'domain_id': domain['id'], 'name': uuid.uuid4().hex}
        }

        with self.test_client() as c:
            c.post(
                '/v3/users',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_users_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_non_existent_user_forbidden(self):
        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                f'/v3/users/{uuid.uuid4().hex}',
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_users_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_users_in_other_domains(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_non_existent_user_forbidden(self):
        with self.test_client() as c:
            c.delete(
                f'/v3/users/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _SystemUserTests,
    _SystemMemberAndReaderUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _SystemUserTests,
    _SystemMemberAndReaderUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _SystemUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_users(self):
        create = {
            'user': {
                'name': uuid.uuid4().hex,
                'domain': CONF.identity.default_domain_id,
            }
        }

        with self.test_client() as c:
            c.post('/v3/users', json=create, headers=self.headers)

    def test_user_can_update_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
            )

    def test_user_cannot_update_non_existent_user_not_found(self):
        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                f'/v3/users/{uuid.uuid4().hex}',
                json=update,
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND,
            )

    def test_user_can_delete_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.delete('/v3/users/{}'.format(user['id']), headers=self.headers)

    def test_user_cannot_delete_non_existent_user_not_found(self):
        with self.test_client() as c:
            c.delete(
                f'/v3/users/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND,
            )


class DomainReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _DomainUserTests,
    _DomainMemberAndReaderUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_reader = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_reader)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_reader['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _DomainUserTests,
    _DomainMemberAndReaderUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_user = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_user)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_user['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _DomainUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(lbragstad): Remove this once the deprecated policies in
        # keystone.common.policies.users have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will apply a logical OR to deprecated policies with
        # new policies to maintain compatibility and give operators a chance to
        # update permissions or update policies without breaking users. This
        # will cause these specific tests to fail since we're trying to correct
        # this broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:get_user': up.SYSTEM_READER_OR_DOMAIN_READER_OR_USER,
                'identity:list_users': up.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:create_user': up.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:update_user': up.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:delete_user': up.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
            }
            f.write(jsonutils.dumps(overridden_policies))

    def test_user_can_create_users_within_domain(self):
        create = {
            'user': {'domain_id': self.domain_id, 'name': uuid.uuid4().hex}
        }

        with self.test_client() as c:
            c.post('/v3/users', json=create, headers=self.headers)

    def test_user_cannot_create_users_within_domain_hyphened_domain_id(self):
        # Finally, show that we can create a new user without any surprises.
        # But if we specify a 'domain-id' instead of a 'domain_id', we get a
        # Forbidden response because we fail a policy check before
        # normalization occurs.
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        create = {
            'user': {'domain-id': domain['id'], 'name': uuid.uuid4().hex}
        }

        with self.test_client() as c:
            c.post(
                '/v3/users',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_create_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        create = {
            'user': {'domain_id': domain['id'], 'name': uuid.uuid4().hex}
        }

        with self.test_client() as c:
            c.post(
                '/v3/users',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_update_users_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
            )

    def test_user_can_update_users_within_domain_hyphened_domain_id(self):
        # If we try updating the user's 'domain_id' by specifying a
        # 'domain-id', then it'll be stored into extras rather than normalized,
        # and the user's actual 'domain_id' is not affected.
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        update = {'user': {'domain-id': domain['id']}}
        with self.test_client() as c:
            r = c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
            )
            self.assertEqual(domain['id'], r.json['user']['domain-id'])
            self.assertEqual(self.domain_id, r.json['user']['domain_id'])

    def test_user_cannot_update_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/users/{}'.format(user['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_non_existent_user_forbidden(self):
        update = {'user': {'email': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                f'/v3/users/{uuid.uuid4().hex}',
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_delete_users_within_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.delete('/v3/users/{}'.format(user['id']), headers=self.headers)

    def test_user_cannot_delete_users_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.delete(
                '/v3/users/{}'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_non_existent_user_forbidden(self):
        with self.test_client() as c:
            c.delete(
                f'/v3/users/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class ProjectReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _ProjectUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project = unit.new_project_ref(domain_id=self.domain_id)
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        self.project_id = project['id']

        project_reader = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(project_reader)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_reader['password'],
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _ProjectUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project = unit.new_project_ref(domain_id=self.domain_id)
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        self.project_id = project['id']

        project_member = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(project_member)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_member['password'],
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _CommonUserTests,
    _ProjectUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.bootstrapper.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(lbragstad): Remove this once the deprecated policies in
        # keystone.common.policies.users have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will OR deprecated policies with new policies to
        # maintain compatibility and give operators a chance to update
        # permissions or update policies without breaking users. This will
        # cause these specific tests to fail since we're trying to correct this
        # broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:get_user': up.SYSTEM_READER_OR_DOMAIN_READER_OR_USER,
                'identity:list_users': up.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:create_user': up.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:update_user': up.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:delete_user': up.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
            }
            f.write(jsonutils.dumps(overridden_policies))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from oslo_config import cfg
from oslo_db import api as oslo_db_api
from oslo_utils import timeutils
import sqlalchemy

from keystone.common import provider_api
from keystone.common import sql
from keystone import exception
from keystone.identity.backends import base as identity_base
from keystone.identity.backends import sql_model as model
from keystone.identity.shadow_backends import base

CONF = cfg.CONF
PROVIDERS = provider_api.ProviderAPIs


class ShadowUsers(base.ShadowUsersDriverBase):
    @sql.handle_conflicts(conflict_type='federated_user')
    def create_federated_user(self, domain_id, federated_dict, email=None):
        local_entity = {
            'domain_id': domain_id,
            'local_id': federated_dict['unique_id'],
            'entity_type': 'user',
        }

        public_id = PROVIDERS.id_generator_api.generate_public_ID(local_entity)

        user = {'id': public_id, 'domain_id': domain_id, 'enabled': True}
        if email:
            user['email'] = email
        with sql.session_for_write() as session:
            federated_ref = model.FederatedUser.from_dict(federated_dict)
            user_ref = model.User.from_dict(user)
            user_ref.created_at = timeutils.utcnow()
            user_ref.federated_users.append(federated_ref)
            session.add(user_ref)
            return identity_base.filter_user(user_ref.to_dict())

    @sql.handle_conflicts(conflict_type='federated_user')
    def create_federated_object(self, fed_dict):
        with sql.session_for_write() as session:
            fed_ref = model.FederatedUser.from_dict(fed_dict)
            session.add(fed_ref)

    def delete_federated_object(self, user_id):
        with sql.session_for_write() as session:
            q = session.query(model.FederatedUser)
            q = q.filter(model.FederatedUser.user_id == user_id)
            q.delete(False)

    def get_federated_objects(self, user_id):
        with sql.session_for_read() as session:
            query = session.query(model.FederatedUser)
            query = query.filter(model.FederatedUser.user_id == user_id)
            fed_ref = []
            for row in query:
                m = model.FederatedUser(
                    idp_id=row.idp_id,
                    protocol_id=row.protocol_id,
                    unique_id=row.unique_id,
                )
                fed_ref.append(m.to_dict())
            return base.federated_objects_to_list(fed_ref)

    def _update_query_with_federated_statements(self, hints, query):
        statements = []
        for filter_ in hints.filters:
            if filter_['name'] == 'idp_id':
                statements.append(
                    model.FederatedUser.idp_id == filter_['value']
                )
            if filter_['name'] == 'protocol_id':
                statements.append(
                    model.FederatedUser.protocol_id == filter_['value']
                )
            if filter_['name'] == 'unique_id':
                statements.append(
                    model.FederatedUser.unique_id == filter_['value']
                )

        # Remove federated attributes to prevent redundancies from
        # sql.filter_limit_query which filters remaining hints
        hints.filters = [
            x
            for x in hints.filters
            if x['name'] not in ('idp_id', 'protocol_id', 'unique_id')
        ]
        if statements:
            query = query.filter(sqlalchemy.and_(*statements))
        return query

    def get_federated_users(self, hints):
        with sql.session_for_read() as session:
            query = (
                session.query(model.User)
                .outerjoin(model.LocalUser)
                .outerjoin(model.FederatedUser)
            )
            query = query.filter(model.User.id == model.FederatedUser.user_id)
            query = self._update_query_with_federated_statements(hints, query)
            name_filter = None
            for filter_ in hints.filters:
                if filter_['name'] == 'name':
                    name_filter = filter_
                    query = query.filter(
                        model.FederatedUser.display_name
                        == name_filter['value']
                    )
                    break
            if name_filter:
                hints.filters.remove(name_filter)
            user_refs = sql.filter_limit_query(model.User, query, hints)
            return [identity_base.filter_user(x.to_dict()) for x in user_refs]

    def get_federated_user(self, idp_id, protocol_id, unique_id):
        # NOTE(notmorgan): Open a session here to ensure .to_dict is called
        # within an active session context. This will prevent lazy-load
        # relationship failure edge-cases
        # FIXME(notmorgan): Eventually this should not call `to_dict` here and
        # rely on something already in the session context to perform the
        # `to_dict` call.
        with sql.session_for_read():
            user_ref = self._get_federated_user(idp_id, protocol_id, unique_id)
            return identity_base.filter_user(user_ref.to_dict())

    def _get_federated_user(self, idp_id, protocol_id, unique_id):
        """Return the found user for the federated identity.

        :param idp_id: The identity provider ID
        :param protocol_id: The federation protocol ID
        :param unique_id: The user's unique ID (unique within the IdP)
        :returns User: Returns a reference to the User

        """
        with sql.session_for_read() as session:
            query = session.query(model.User).outerjoin(model.LocalUser)
            query = query.join(model.FederatedUser)
            query = query.filter(model.FederatedUser.idp_id == idp_id)
            query = query.filter(
                model.FederatedUser.protocol_id == protocol_id
            )
            query = query.filter(model.FederatedUser.unique_id == unique_id)
            try:
                user_ref = query.one()
            except sql.NotFound:
                raise exception.UserNotFound(user_id=unique_id)
            return user_ref

    def set_last_active_at(self, user_id):
        with sql.session_for_write() as session:
            user_ref = session.get(model.User, user_id)
            if user_ref:
                user_ref.last_active_at = timeutils.utcnow().date()

    @sql.handle_conflicts(conflict_type='federated_user')
    def update_federated_user_display_name(
        self, idp_id, protocol_id, unique_id, display_name
    ):
        with sql.session_for_write() as session:
            query = session.query(model.FederatedUser)
            query = query.filter(model.FederatedUser.idp_id == idp_id)
            query = query.filter(
                model.FederatedUser.protocol_id == protocol_id
            )
            query = query.filter(model.FederatedUser.unique_id == unique_id)
            query = query.filter(
                model.FederatedUser.display_name != display_name
            )
            query.update({'display_name': display_name})
            return

    @sql.handle_conflicts(conflict_type='nonlocal_user')
    def create_nonlocal_user(self, user_dict):
        new_user_dict = copy.deepcopy(user_dict)
        # remove local_user attributes from new_user_dict
        new_user_dict.pop('name', None)
        new_user_dict.pop('password', None)
        # create nonlocal_user dict
        new_nonlocal_user_dict = {'name': user_dict['name']}
        with sql.session_for_write() as session:
            new_nonlocal_user_ref = model.NonLocalUser.from_dict(
                new_nonlocal_user_dict
            )
            new_user_ref = model.User.from_dict(new_user_dict)
            new_user_ref.created_at = timeutils.utcnow()
            new_user_ref.nonlocal_user = new_nonlocal_user_ref
            session.add(new_user_ref)
            return identity_base.filter_user(new_user_ref.to_dict())

    @oslo_db_api.wrap_db_retry(retry_on_deadlock=True)
    def delete_user(self, user_id):
        with sql.session_for_write() as session:
            ref = self._get_user(session, user_id)

            q = session.query(model.UserGroupMembership)
            q = q.filter_by(user_id=user_id)
            q.delete(False)

            session.delete(ref)

    def get_user(self, user_id):
        with sql.session_for_read() as session:
            user_ref = self._get_user(session, user_id)
            return identity_base.filter_user(user_ref.to_dict())

    def _get_user(self, session, user_id):
        user_ref = session.get(model.User, user_id)
        if not user_ref:
            raise exception.UserNotFound(user_id=user_id)
        return user_ref

    def list_federated_users_info(self, hints=None):
        with sql.session_for_read() as session:
            query = session.query(model.FederatedUser)
            fed_user_refs = sql.filter_limit_query(
                model.FederatedUser, query, hints
            )
            return [x.to_dict() for x in fed_user_refs]

    def add_user_to_group_expires(self, user_id, group_id):
        def get_federated_user():
            with sql.session_for_read() as session:
                query = session.query(model.FederatedUser)
                query = query.filter_by(user_id=user_id)
                user = query.first()
                if not user:
                    # Note(knikolla): This shouldn't really ever happen, since
                    # this requires the user to already be logged in.
                    raise exception.UserNotFound(user_id=user_id)
                return user

        with sql.session_for_write() as session:
            user = get_federated_user()
            query = session.query(model.ExpiringUserGroupMembership)
            query = query.filter_by(user_id=user_id)
            query = query.filter_by(group_id=group_id)
            membership = query.first()

            if membership:
                membership.last_verified = timeutils.utcnow()
            else:
                session.add(
                    model.ExpiringUserGroupMembership(
                        user_id=user_id,
                        group_id=group_id,
                        idp_id=user.idp_id,
                        last_verified=timeutils.utcnow(),
                    )
                )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(morgan): Import relevant stuff so importing individual under-pinnings
# isn't needed, keystone.server.flask exposes all the interesting bits
# needed to develop restful APIs for keystone.

from keystone.server.flask.common import APIBase  # noqa
from keystone.server.flask.common import base_url  # noqa
from keystone.server.flask.common import construct_json_home_data  # noqa
from keystone.server.flask.common import construct_resource_map  # noqa
from keystone.server.flask.common import full_url  # noqa
from keystone.server.flask.common import JsonHomeData  # noqa
from keystone.server.flask.common import ResourceBase  # noqa
from keystone.server.flask.common import ResourceMap  # noqa
from keystone.server.flask.common import unenforced_api  # noqa


# NOTE(morgan): This allows for from keystone.flask import * and have all the
# cool stuff needed to develop new APIs within a module/subsystem
__all__ = (
    'APIBase',
    'JsonHomeData',
    'ResourceBase',
    'ResourceMap',
    'base_url',
    'construct_json_home_data',
    'construct_resource_map',
    'full_url',
    'unenforced_api',
)



# Copyright 2011 Piston Cloud Computing, Inc.
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import subprocess
from unittest import mock
import uuid

from oslo_policy import policy as common_policy

from keystone.common import policies
from keystone.common.rbac_enforcer import policy
import keystone.conf
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF


class PolicyFileTestCase(unit.TestCase):
    def setUp(self):
        # self.tmpfilename should exist before setUp super is called
        # this is to ensure it is available for the config_fixture in
        # the config_overrides call.
        self.tempfile = self.useFixture(temporaryfile.SecureTempFile())
        self.tmpfilename = self.tempfile.file_name
        super().setUp()
        self.target = {}

    def _policy_fixture(self):
        return ksfixtures.Policy(
            self.config_fixture, policy_file=self.tmpfilename
        )

    def test_modified_policy_reloads(self):
        action = "example:test"
        empty_credentials = {}
        with open(self.tmpfilename, "w") as policyfile:
            policyfile.write("""{"example:test": []}""")
        policy.enforce(empty_credentials, action, self.target)
        with open(self.tmpfilename, "w") as policyfile:
            policyfile.write("""{"example:test": ["false:false"]}""")
        policy._ENFORCER._enforcer.clear()
        self.assertRaises(
            exception.ForbiddenAction,
            policy.enforce,
            empty_credentials,
            action,
            self.target,
        )


class PolicyTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.rules = {
            "true": [],
            "example:allowed": [],
            "example:denied": [["false:false"]],
            "example:get_http": [["http:http://www.example.com"]],
            "example:my_file": [
                ["role:compute_admin"],
                ["project_id:%(project_id)s"],
            ],
            "example:early_and_fail": [["false:false", "rule:true"]],
            "example:early_or_success": [["rule:true"], ["false:false"]],
            "example:lowercase_admin": [["role:admin"], ["role:sysadmin"]],
            "example:uppercase_admin": [["role:ADMIN"], ["role:sysadmin"]],
        }

        # NOTE(vish): then overload underlying policy engine
        self._set_rules()
        self.credentials = {}
        self.target = {}

    def _set_rules(self):
        these_rules = common_policy.Rules.from_dict(self.rules)
        policy._ENFORCER._enforcer.set_rules(these_rules)

    def test_enforce_nonexistent_action_throws(self):
        action = "example:noexist"
        self.assertRaises(
            exception.ForbiddenAction,
            policy.enforce,
            self.credentials,
            action,
            self.target,
        )

    def test_enforce_bad_action_throws(self):
        action = "example:denied"
        self.assertRaises(
            exception.ForbiddenAction,
            policy.enforce,
            self.credentials,
            action,
            self.target,
        )

    def test_enforce_good_action(self):
        action = "example:allowed"
        policy.enforce(self.credentials, action, self.target)

    def test_templatized_enforcement(self):
        target_mine = {'project_id': 'fake'}
        target_not_mine = {'project_id': 'another'}
        credentials = {'project_id': 'fake', 'roles': []}
        action = "example:my_file"
        policy.enforce(credentials, action, target_mine)
        self.assertRaises(
            exception.ForbiddenAction,
            policy.enforce,
            credentials,
            action,
            target_not_mine,
        )

    def test_early_AND_enforcement(self):
        action = "example:early_and_fail"
        self.assertRaises(
            exception.ForbiddenAction,
            policy.enforce,
            self.credentials,
            action,
            self.target,
        )

    def test_early_OR_enforcement(self):
        action = "example:early_or_success"
        policy.enforce(self.credentials, action, self.target)

    def test_ignore_case_role_check(self):
        lowercase_action = "example:lowercase_admin"
        uppercase_action = "example:uppercase_admin"
        # NOTE(dprince): We mix case in the Admin role here to ensure
        # case is ignored
        admin_credentials = {'roles': ['AdMiN']}
        policy.enforce(admin_credentials, lowercase_action, self.target)
        policy.enforce(admin_credentials, uppercase_action, self.target)


class PolicyScopeTypesEnforcementTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        rule = common_policy.RuleDefault(
            name='foo', check_str='', scope_types=['system']
        )
        policy._ENFORCER._enforcer.register_default(rule)
        self.credentials = {}
        self.action = 'foo'
        self.target = {}

    def test_forbidden_is_raised_if_enforce_scope_is_true(self):
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)
        self.assertRaises(
            exception.ForbiddenAction,
            policy.enforce,
            self.credentials,
            self.action,
            self.target,
        )

    def test_warning_message_is_logged_if_enforce_scope_is_false(self):
        self.config_fixture.config(group='oslo_policy', enforce_scope=False)
        expected_msg = (
            'Policy "foo": "" failed scope check. The token used to make the '
            'request was project scoped but the policy requires [\'system\'] '
            'scope. This behavior may change in the future where using the '
            'intended scope is required'
        )
        with mock.patch('warnings.warn') as mock_warn:
            policy.enforce(self.credentials, self.action, self.target)
            mock_warn.assert_called_with(expected_msg)


class PolicyJsonTestCase(unit.TestCase):
    def _get_default_policy_rules(self):
        """Return a dictionary of all in-code policies.

        All policies have a default value that is maintained in code.
        This method returns a dictionary containing all default policies.
        """
        rules = {}
        for rule in policies.list_rules():
            rules[rule.name] = rule.check_str
        return rules

    def test_policies_loads(self):
        action = 'identity:list_projects'
        target = {
            'user_id': uuid.uuid4().hex,
            'user.domain_id': uuid.uuid4().hex,
            'group.domain_id': uuid.uuid4().hex,
            'project.domain_id': uuid.uuid4().hex,
            'project_id': uuid.uuid4().hex,
            'domain_id': uuid.uuid4().hex,
        }
        credentials = {
            'username': uuid.uuid4().hex,
            'token': uuid.uuid4().hex,
            'project_name': None,
            'user_id': uuid.uuid4().hex,
            'roles': ['admin'],
            'is_admin': True,
            'is_admin_project': True,
            'project_id': None,
            'domain_id': uuid.uuid4().hex,
        }

        # The enforcer is setup behind the scenes and registers the in code
        # default policies.
        result = policy._ENFORCER._enforcer.enforce(
            action, target, credentials
        )
        self.assertTrue(result)

    def test_all_targets_documented(self):
        policy_keys = self._get_default_policy_rules()

        # These keys are in the policy.yaml but aren't targets.
        policy_rule_keys = [
            'admin_or_owner',
            'admin_or_token_subject',
            'admin_required',
            'owner',
            'service_admin_or_token_subject',
            'service_or_admin',
            'service_role',
            'token_subject',
            'domain_managed_target_role',
        ]

        def read_doc_targets():
            # Parse the doc/source/policy_mapping.rst file and return the
            # targets.

            doc_path = os.path.join(
                unit.ROOTDIR,
                'doc',
                'source',
                'getting-started',
                'policy_mapping.rst',
            )
            with open(doc_path) as doc_file:
                for line in doc_file:
                    if line.startswith('Target'):
                        break
                for line in doc_file:
                    # Skip === line
                    if line.startswith('==='):
                        break
                for line in doc_file:
                    line = line.rstrip()
                    if not line or line.startswith(' '):
                        continue
                    if line.startswith('=='):
                        break
                    target, dummy, dummy = line.partition(' ')
                    yield str(target)

        doc_targets = list(read_doc_targets())
        self.assertCountEqual(policy_keys, doc_targets + policy_rule_keys)


class GeneratePolicyFileTestCase(unit.TestCase):
    def test_policy_generator_from_command_line(self):
        # This test ensures keystone.common.policy:get_enforcer ignores
        # unexpected arguments before handing them off to oslo.config, which
        # will fail and prevent users from generating policy files.
        ret_val = subprocess.Popen(
            ['oslopolicy-policy-generator', '--namespace', 'keystone'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        output = ret_val.communicate()
        self.assertEqual(ret_val.returncode, 0, output)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import binascii
import os

from cryptography.exceptions import InvalidKey
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from keystone.common import password_hashers
from keystone import exception


class Sha512(password_hashers.PasswordHasher):
    """passlib transition class for PBKDF2 Sha512 password hashing"""

    name: str = "pbkdf2_sha512"
    ident: str = "$pbkdf2-sha512$"
    hash_algo = hashes.SHA512()

    @staticmethod
    def hash(password: bytes, salt_size: int = 16, rounds: int = 25000) -> str:
        """Generate password hash string with ident and params

        https://cryptography.io/en/stable/hazmat/primitives/key-derivation-functions/#pbkdf2

        :param bytes password: Password to be hashed.
        :param bytes salt: Salt.
        :param int iterations: Iterations count

        :returns: String in format
            `$pbkdf2-sha512$ln=logN,r=R,p=P$salt$checksum`
        """
        salt: bytes = os.urandom(salt_size)

        # Prepave the kdf function with params
        kdf = PBKDF2HMAC(
            algorithm=Sha512.hash_algo, length=64, salt=salt, iterations=rounds
        )

        # derive - create a digest
        key: bytes = kdf.derive(password)

        # make a `str` digest compatible with passlib
        digest_str: str = (
            binascii.b2a_base64(key).rstrip(b"=\n").decode("ascii")
        )
        # make a `str` salt
        salt_str: str = (
            binascii.b2a_base64(salt).rstrip(b"=\n").decode("ascii")
        )

        return f"$pbkdf2-sha512${rounds}${salt_str}${digest_str}"

    @staticmethod
    def verify(password: bytes, hashed: str) -> bool:
        """Verify hashing password would be equal to the `hashed` value

        :param bytes password: Password to verify
        :param string hashed: Hashed password. Used to extract hashing
            parameters

        :returns: boolean whether hashing password with the same parameters
            would match hashed value
        """
        data: str = hashed
        # split hashed string to extract parameters
        parts: list[str] = data[1:].split("$")
        rounds: int

        if len(parts) == 4:
            _, rounds_str, salt_str, digest_str = parts
            # Convert salt and digest back to bytes as opposite to how passlib
            # serializes them
            salt: bytes = password_hashers.b64s_decode(
                salt_str.replace(".", "+").encode("ascii")
            )
            digest: bytes = password_hashers.b64s_decode(
                digest_str.replace(".", "+").encode("ascii")
            )
            rounds = int(rounds_str)
        else:
            raise exception.PasswordValidationError("malformed password hash")

        # Prepave the kdf function with params
        kdf = PBKDF2HMAC(
            algorithm=Sha512.hash_algo, length=64, salt=salt, iterations=rounds
        )

        # Verify the key.
        # NOTE(gtema): cryptography raises exception when key does not match
        try:
            kdf.verify(password, digest)
            return True
        except InvalidKey:
            return False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from keystone.common.policies import base

ADMIN_OR_SYSTEM_OR_DOMAIN_OR_PROJECT_USER = (
    base.RULE_ADMIN_REQUIRED + ' or '
    '(' + base.SYSTEM_READER + ') or '
    '('
    'domain_id:%(target.limit.domain.id)s or '
    'domain_id:%(target.limit.project.domain_id)s'
    ') or '
    '('
    'project_id:%(target.limit.project_id)s and not '
    'None:%(target.limit.project_id)s'
    ')'
)

limit_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_limit_model',
        check_str='',
        scope_types=['system', 'domain', 'project'],
        description='Get limit enforcement model.',
        operations=[
            {'path': '/v3/limits/model', 'method': 'GET'},
            {'path': '/v3/limits/model', 'method': 'HEAD'},
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_limit',
        check_str=ADMIN_OR_SYSTEM_OR_DOMAIN_OR_PROJECT_USER,
        scope_types=['system', 'domain', 'project'],
        description='Show limit details.',
        operations=[
            {'path': '/v3/limits/{limit_id}', 'method': 'GET'},
            {'path': '/v3/limits/{limit_id}', 'method': 'HEAD'},
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_limits',
        check_str='',
        scope_types=['system', 'domain', 'project'],
        description='List limits.',
        operations=[
            {'path': '/v3/limits', 'method': 'GET'},
            {'path': '/v3/limits', 'method': 'HEAD'},
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_limits',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create limits.',
        operations=[{'path': '/v3/limits', 'method': 'POST'}],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_limit',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Update limit.',
        operations=[{'path': '/v3/limits/{limit_id}', 'method': 'PATCH'}],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_limit',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete limit.',
        operations=[{'path': '/v3/limits/{limit_id}', 'method': 'DELETE'}],
    ),
]


def list_rules():
    return limit_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import validation
from keystone.common.validation import parameter_types
from keystone import exception
from keystone.i18n import _

token_issue = {
    'type': 'object',
    'properties': {
        'identity': {
            'type': 'object',
            'properties': {
                'methods': {'type': 'array', 'items': {'type': 'string'}},
                'password': {
                    'type': 'object',
                    'properties': {
                        'user': {
                            'type': 'object',
                            'properties': {
                                'id': {'type': 'string'},
                                'name': {'type': 'string'},
                                'password': {'type': 'string'},
                                'domain': {
                                    'type': 'object',
                                    'properties': {
                                        'id': {'type': 'string'},
                                        'name': {'type': 'string'},
                                    },
                                },
                            },
                        }
                    },
                },
                'token': {
                    'type': 'object',
                    'properties': {'id': {'type': 'string'}},
                    'required': ['id'],
                },
            },
            'required': ['methods'],
        },
        'scope': {
            # For explicit unscoped authentication the type should not be
            # strictly string. Although keystone server specifies the value
            # to be 'unscoped', old versions of keystoneauth might still be
            # using `"scope": {'unscoped': {}}` instead of
            # `"scope": "unscoped"`
            # https://bugs.launchpad.net/keystoneauth/+bug/1637682/
            'type': ['object', 'string'],
            'properties': {
                'project': {
                    'type': 'object',
                    'properties': {
                        'name': {'type': 'string'},
                        'id': {'type': 'string'},
                        'domain': {
                            'type': 'object',
                            'properties': {
                                'id': {'type': 'string'},
                                'name': {'type': 'string'},
                            },
                        },
                    },
                },
                'domain': {
                    'type': 'object',
                    'properties': {
                        'id': {'type': 'string'},
                        'name': {'type': 'string'},
                    },
                },
                'OS-TRUST:trust': {
                    'type': 'object',
                    'properties': {'id': {'type': 'string'}},
                },
                'system': {
                    'type': 'object',
                    'properties': {'all': parameter_types.boolean},
                },
            },
        },
    },
    'required': ['identity'],
}


def validate_issue_token_auth(auth=None):
    if auth is None:
        return
    validation.lazy_validate(token_issue, auth)

    user = auth['identity'].get('password', {}).get('user')
    if user is not None:
        if 'id' not in user and 'name' not in user:
            msg = _(
                'Invalid input for field identity/password/user: '
                'id or name must be present.'
            )
            raise exception.SchemaValidationError(detail=msg)

        domain = user.get('domain')
        if domain is not None:
            if 'id' not in domain and 'name' not in domain:
                msg = _(
                    'Invalid input for field identity/password/user/domain: '
                    'id or name must be present.'
                )
                raise exception.SchemaValidationError(detail=msg)

    scope = auth.get('scope')
    if scope is not None and isinstance(scope, dict):
        project = scope.get('project')
        if project is not None:
            if 'id' not in project and 'name' not in project:
                msg = _(
                    'Invalid input for field scope/project: '
                    'id or name must be present.'
                )
                raise exception.SchemaValidationError(detail=msg)
            domain = project.get('domain')
            if domain is not None:
                if 'id' not in domain and 'name' not in domain:
                    msg = _(
                        'Invalid input for field scope/project/domain: '
                        'id or name must be present.'
                    )
                    raise exception.SchemaValidationError(detail=msg)
        domain = scope.get('domain')
        if domain is not None:
            if 'id' not in domain and 'name' not in domain:
                msg = _(
                    'Invalid input for field scope/domain: '
                    'id or name must be present.'
                )
                raise exception.SchemaValidationError(detail=msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# LOG some debug output about the request. This was originally in the
# dispatch middleware

import flask
from oslo_log import log

LOG = log.getLogger(__name__)


def log_request_info():
    # Add in any extra debug logging about the request that is desired
    # note that this is executed prior to routing the request to a resource
    # so the data is somewhat raw.
    LOG.debug('REQUEST_METHOD: `%s`', flask.request.method)
    LOG.debug('SCRIPT_NAME: `%s`', flask.request.script_root)
    LOG.debug('PATH_INFO: `%s`', flask.request.path)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import keystone.conf

CONF = keystone.conf.CONF


def symptom_database_connection_is_not_SQLite():
    """SQLite is not recommended for production deployments.

    SQLite does not enforce type checking and has limited support for
    migrations, making it unsuitable for use in keystone. Please change your
    `keystone.conf [database] connection` value to point to a supported
    database driver, such as MySQL.
    """  # noqa: D403
    return (
        CONF.database.connection is not None
        and 'sqlite' in CONF.database.connection
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The domain API is now aware of system scope and default roles."
)

deprecated_list_domains = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_domains',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_get_domain = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_domain',
    check_str=base.RULE_ADMIN_OR_TARGET_DOMAIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_update_domain = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_domain',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_create_domain = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_domain',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_delete_domain = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_domain',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
ADMIN_OR_SYSTEM_USER_OR_DOMAIN_USER_OR_PROJECT_USER = (
    base.RULE_ADMIN_REQUIRED + ' or '
    '(role:reader and system_scope:all) or '
    'token.domain.id:%(target.domain.id)s or '
    'token.project.domain.id:%(target.domain.id)s'
)
ADMIN_OR_SYSTEM_READER_OR_DOMAIN_READER = (
    base.RULE_ADMIN_OR_SYSTEM_READER + ' or '
    '(role:reader and domain_id:%(target.domain.id)s)'
)


domain_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_domain',
        # NOTE(lbragstad): This policy allows system, domain, and
        # project-scoped tokens.
        check_str=ADMIN_OR_SYSTEM_USER_OR_DOMAIN_USER_OR_PROJECT_USER,
        scope_types=['system', 'domain', 'project'],
        description='Show domain details.',
        operations=[{'path': '/v3/domains/{domain_id}', 'method': 'GET'}],
        deprecated_rule=deprecated_get_domain,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_domains',
        check_str=ADMIN_OR_SYSTEM_READER_OR_DOMAIN_READER,
        scope_types=['system', 'domain', 'project'],
        description='List domains.',
        operations=[{'path': '/v3/domains', 'method': 'GET'}],
        deprecated_rule=deprecated_list_domains,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_domain',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create domain.',
        operations=[{'path': '/v3/domains', 'method': 'POST'}],
        deprecated_rule=deprecated_create_domain,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_domain',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Update domain.',
        operations=[{'path': '/v3/domains/{domain_id}', 'method': 'PATCH'}],
        deprecated_rule=deprecated_update_domain,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_domain',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete domain.',
        operations=[{'path': '/v3/domains/{domain_id}', 'method': 'DELETE'}],
        deprecated_rule=deprecated_delete_domain,
    ),
]


def list_rules():
    return domain_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import keystone.conf

CONF = keystone.conf.CONF


def symptom_comma_in_SAML_public_certificate_path():
    """`[saml] certfile` should not contain a comma (`,`).

    Because a comma is part of the API between keystone and the external
    xmlsec1 binary which utilizes the certificate, keystone cannot include a
    comma in the path to the public certificate file.
    """
    return ',' in CONF.saml.certfile


def symptom_comma_in_SAML_private_key_file_path():
    """`[saml] certfile` should not contain a comma (`,`).

    Because a comma is part of the API between keystone and the external
    xmlsec1 binary which utilizes the key, keystone cannot include a comma in
    the path to the private key file.
    """
    return ',' in CONF.saml.keyfile



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Tests for database migrations for the database.

These are "opportunistic" tests which allow testing against all three databases
(sqlite in memory, mysql, pg) in a properly configured unit test environment.

For the opportunistic testing you need to set up DBs named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost. The
test will then use that DB and username/password combo to run the tests.
"""

import fixtures
from oslo_db import options as db_options
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import test_migrations
from oslo_log.fixture import logging_error as log_fixture
from oslo_log import log as logging
from oslotest import base

# We need to import all of these so the tables are registered. It would be
# easier if these were all in a central location :(
import keystone.application_credential.backends.sql  # noqa: F401
import keystone.assignment.backends.sql  # noqa: F401
import keystone.assignment.role_backends.sql_model  # noqa: F401
import keystone.catalog.backends.sql  # noqa: F401
from keystone.common import sql
from keystone.common.sql import upgrades
import keystone.conf
import keystone.credential.backends.sql  # noqa: F401
import keystone.endpoint_policy.backends.sql  # noqa: F401
import keystone.federation.backends.sql  # noqa: F401
import keystone.identity.backends.sql_model  # noqa: F401
import keystone.identity.mapping_backends.sql  # noqa: F401
import keystone.limit.backends.sql  # noqa: F401
import keystone.oauth1.backends.sql  # noqa: F401
import keystone.policy.backends.sql  # noqa: F401
import keystone.resource.backends.sql_model  # noqa: F401
import keystone.resource.config_backends.sql  # noqa: F401
import keystone.revoke.backends.sql  # noqa: F401
from keystone.tests.unit import ksfixtures
import keystone.trust.backends.sql  # noqa: F401

CONF = keystone.conf.CONF
LOG = logging.getLogger(__name__)


class KeystoneModelsMigrationsSync(test_migrations.ModelsMigrationsSync):
    """Test sqlalchemy-migrate migrations."""

    # Migrations can take a long time, particularly on underpowered CI nodes.
    # Give them some breathing room.
    TIMEOUT_SCALING_FACTOR = 4

    def setUp(self):
        # Ensure BaseTestCase's ConfigureLogging fixture is disabled since
        # we're using our own (StandardLogging).
        with fixtures.EnvironmentVariable('OS_LOG_CAPTURE', '0'):
            super().setUp()

        self.useFixture(log_fixture.get_logging_handle_error_fixture())
        self.useFixture(ksfixtures.WarningsFixture())
        self.useFixture(ksfixtures.StandardLogging())

        self.engine = enginefacade.writer.get_engine()

        # Configure our connection string in CONF and enable SQLite fkeys
        db_options.set_defaults(CONF, connection=self.engine.url)

        # TODO(stephenfin): Do we need this? I suspect not since we're using
        # enginefacade.write.get_engine() directly above
        # Override keystone's context manager to be oslo.db's global context
        # manager.
        sql.core._TESTING_USE_GLOBAL_CONTEXT_MANAGER = True
        self.addCleanup(
            setattr, sql.core, '_TESTING_USE_GLOBAL_CONTEXT_MANAGER', False
        )
        self.addCleanup(sql.cleanup)

    def db_sync(self, engine):
        upgrades.offline_sync_database_to_version(engine=engine)

    def get_engine(self):
        return self.engine

    def get_metadata(self):
        return sql.ModelBase.metadata

    def include_object(self, object_, name, type_, reflected, compare_to):
        if type_ == 'table':
            # migrate_version is a sqlalchemy-migrate control table and
            # isn't included in the models
            if name == 'migrate_version':
                return False

            # This is created in tests and isn't a "real" table
            if name == 'test_table':
                return False

            # FIXME(stephenfin): This was dropped in commit 93aff6e42 but the
            # migrations were never adjusted
            if name == 'token':
                return False

        return True

    def filter_metadata_diff(self, diff):
        """Filter changes before assert in test_models_sync().

        :param diff: a list of differences (see `compare_metadata()` docs for
            details on format)
        :returns: a list of differences
        """
        new_diff = []
        for element in diff:
            # The modify_foo elements are lists; everything else is a tuple
            if isinstance(element, list):
                if element[0][0] == 'modify_nullable':
                    if (element[0][2], element[0][3]) in (
                        ('credential', 'encrypted_blob'),
                        ('credential', 'key_hash'),
                        ('federated_user', 'user_id'),
                        ('federated_user', 'idp_id'),
                        ('local_user', 'user_id'),
                        ('nonlocal_user', 'user_id'),
                        ('password', 'local_user_id'),
                    ):
                        continue  # skip

                if element[0][0] == 'modify_default':
                    if (element[0][2], element[0][3]) in (
                        ('password', 'created_at_int'),
                        ('password', 'self_service'),
                        ('project', 'is_domain'),
                        ('service_provider', 'relay_state_prefix'),
                    ):
                        continue  # skip
            else:
                # FIXME(stephenfin): These indexes are present in the
                # migrations but not on the equivalent models. Resolve by
                # updating the models.
                if element[0] == 'add_index':
                    if (
                        element[1].table.name,
                        [x.name for x in element[1].columns],
                    ) in (
                        ('access_rule', ['external_id']),
                        ('access_rule', ['user_id']),
                        ('revocation_event', ['revoked_at']),
                        ('system_assignment', ['actor_id']),
                        ('user', ['default_project_id']),
                    ):
                        continue  # skip

                # FIXME(stephenfin): These indexes are present on the models
                # but not in the migrations. Resolve by either removing from
                # the models or adding new migrations.
                if element[0] == 'remove_index':
                    if (
                        element[1].table.name,
                        [x.name for x in element[1].columns],
                    ) in (
                        ('access_rule', ['external_id']),
                        ('access_rule', ['user_id']),
                        ('access_token', ['consumer_id']),
                        ('endpoint', ['service_id']),
                        ('revocation_event', ['revoked_at']),
                        ('user', ['default_project_id']),
                        ('user_group_membership', ['group_id']),
                        (
                            'trust',
                            [
                                'trustor_user_id',
                                'trustee_user_id',
                                'project_id',
                                'impersonation',
                                'expires_at',
                                'expires_at_int',
                            ],
                        ),
                        (),
                    ):
                        continue  # skip

                # FIXME(stephenfin): These fks are present in the
                # migrations but not on the equivalent models. Resolve by
                # updating the models.
                if element[0] == 'add_fk':
                    if (element[1].table.name, element[1].column_keys) in (
                        (
                            'application_credential_access_rule',
                            ['access_rule_id'],
                        ),
                        ('limit', ['registered_limit_id']),
                        ('registered_limit', ['service_id']),
                        ('registered_limit', ['region_id']),
                        ('endpoint', ['region_id']),
                    ):
                        continue  # skip

                # FIXME(stephenfin): These indexes are present on the models
                # but not in the migrations. Resolve by either removing from
                # the models or adding new migrations.
                if element[0] == 'remove_fk':
                    if (element[1].table.name, element[1].column_keys) in (
                        (
                            'application_credential_access_rule',
                            ['access_rule_id'],
                        ),
                        ('endpoint', ['region_id']),
                        ('assignment', ['role_id']),
                    ):
                        continue  # skip

            new_diff.append(element)

        return new_diff


class TestModelsSyncSQLite(
    KeystoneModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    base.BaseTestCase,
):
    pass


class TestModelsSyncMySQL(
    KeystoneModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    base.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestModelsSyncPostgreSQL(
    KeystoneModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    base.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

from oslo_utils import timeutils

from keystone.common.cache import _context_cache
from keystone.common import utils as ks_utils
from keystone import exception
from keystone.models import receipt_model
from keystone.tests.unit import base_classes


class TestReceiptSerialization(base_classes.TestCaseWithBootstrap):
    def setUp(self):
        super().setUp()
        self.admin_user_id = self.bootstrapper.admin_user_id

        self.receipt_id = uuid.uuid4().hex
        issued_at = timeutils.utcnow()
        self.issued_at = ks_utils.isotime(at=issued_at, subsecond=True)

        # Reach into the cache registry and pull out an instance of the
        # _ReceiptModelHandler so that we can interact and test it directly (as
        # opposed to using PROVIDERS or managers to invoke it).
        receipt_handler_id = receipt_model._ReceiptModelHandler.identity
        self.receipt_handler = _context_cache._registry.get(receipt_handler_id)

        self.exp_receipt = receipt_model.ReceiptModel()
        self.exp_receipt.user_id = self.admin_user_id
        self.exp_receipt.mint(self.receipt_id, self.issued_at)

    def test_serialize_and_deserialize_receipt_model(self):
        serialized = self.receipt_handler.serialize(self.exp_receipt)
        receipt = self.receipt_handler.deserialize(serialized)

        self.assertEqual(self.exp_receipt.user_id, receipt.user_id)
        self.assertEqual(self.exp_receipt.id, receipt.id)
        self.assertEqual(self.exp_receipt.issued_at, receipt.issued_at)

    @mock.patch.object(
        receipt_model.ReceiptModel, '__init__', side_effect=Exception
    )
    def test_error_handling_in_deserialize(self, handler_mock):
        serialized = self.receipt_handler.serialize(self.exp_receipt)
        self.assertRaises(
            exception.CacheDeserializationError,
            self.receipt_handler.deserialize,
            serialized,
        )



# Copyright 2018 Huawei
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy

from keystone.common import cache
from keystone.common import driver_hints
from keystone.common import manager
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.limit.models import base

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs

MEMOIZE = cache.get_memoization_decorator(group='unified_limit')


class Manager(manager.Manager):
    driver_namespace = 'keystone.unified_limit'
    _provides_api = 'unified_limit_api'

    def __init__(self):
        unified_limit_driver = CONF.unified_limit.driver
        super().__init__(unified_limit_driver)

        self.enforcement_model = base.load_driver(
            CONF.unified_limit.enforcement_model
        )

    def check_project_depth(self):
        """Check if project depth satisfies current enforcement model."""
        PROVIDERS.resource_api.check_project_depth(
            self.enforcement_model.MAX_PROJECT_TREE_DEPTH
        )

    def _assert_resource_exist(self, unified_limit, target):
        try:
            service_id = unified_limit.get('service_id')
            if service_id is not None:
                PROVIDERS.catalog_api.get_service(service_id)
            region_id = unified_limit.get('region_id')
            if region_id is not None:
                PROVIDERS.catalog_api.get_region(region_id)
            project_id = unified_limit.get('project_id')
            if project_id is not None:
                project = PROVIDERS.resource_api.get_project(project_id)
                if project['is_domain']:
                    # Treat the input limit as domain level limit.
                    unified_limit['domain_id'] = unified_limit.pop(
                        'project_id'
                    )
            domain_id = unified_limit.get('domain_id')
            if domain_id is not None:
                PROVIDERS.resource_api.get_domain(domain_id)

        except exception.ServiceNotFound:
            raise exception.ValidationError(
                attribute='service_id', target=target
            )
        except exception.RegionNotFound:
            raise exception.ValidationError(
                attribute='region_id', target=target
            )
        except exception.ProjectNotFound:
            raise exception.ValidationError(
                attribute='project_id', target=target
            )
        except exception.DomainNotFound:
            raise exception.ValidationError(
                attribute='domain_id', target=target
            )

    def get_model(self):
        """Return information of the configured enforcement model."""
        return {
            'name': self.enforcement_model.NAME,
            'description': self.enforcement_model.DESCRIPTION,
        }

    def create_registered_limits(self, registered_limits):
        for registered_limit in registered_limits:
            self._assert_resource_exist(registered_limit, 'registered_limit')
        return self.driver.create_registered_limits(registered_limits)

    def update_registered_limit(self, registered_limit_id, registered_limit):
        self._assert_resource_exist(registered_limit, 'registered_limit')
        updated_registered_limit = self.driver.update_registered_limit(
            registered_limit_id, registered_limit
        )
        self.get_registered_limit.invalidate(
            self, updated_registered_limit['id']
        )
        return updated_registered_limit

    @manager.response_truncated
    def list_registered_limits(self, hints=None):
        return self.driver.list_registered_limits(
            hints or driver_hints.Hints()
        )

    @MEMOIZE
    def get_registered_limit(self, registered_limit_id):
        return self.driver.get_registered_limit(registered_limit_id)

    def delete_registered_limit(self, registered_limit_id):
        self.driver.delete_registered_limit(registered_limit_id)
        self.get_registered_limit.invalidate(self, registered_limit_id)

    def create_limits(self, limits):
        for limit in limits:
            self._assert_resource_exist(limit, 'limit')
        self.enforcement_model.check_limit(copy.deepcopy(limits))
        return self.driver.create_limits(limits)

    def update_limit(self, limit_id, limit):
        self._assert_resource_exist(limit, 'limit')
        limit_ref = self.get_limit(limit_id)
        limit_ref.update(limit)
        self.enforcement_model.check_limit(copy.deepcopy([limit_ref]))
        updated_limit = self.driver.update_limit(limit_id, limit)
        self.get_limit.invalidate(self, updated_limit['id'])
        return updated_limit

    @manager.response_truncated
    def list_limits(self, hints=None):
        return self.driver.list_limits(hints or driver_hints.Hints())

    @MEMOIZE
    def get_limit(self, limit_id):
        return self.driver.get_limit(limit_id)

    def delete_limit(self, limit_id):
        self.driver.delete_limit(limit_id)
        self.get_limit.invalidate(self, limit_id)

    def delete_limits_for_project(self, project_id):
        limit_ids = self.driver.delete_limits_for_project(project_id)
        for limit_id in limit_ids:
            self.get_limit.invalidate(self, limit_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from dogpile.cache import api as dogpile
from dogpile.cache.backends import memory
from oslo_config import fixture as config_fixture

from keystone.common import cache
import keystone.conf
from keystone.tests import unit

CONF = keystone.conf.CONF


class TestCacheRegion(unit.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.config_fixture = self.useFixture(config_fixture.Config(CONF))
        self.config_fixture.config(
            # TODO(morganfainberg): Make Cache Testing a separate test case
            # in tempest, and move it out of the base unit tests.
            group='cache',
            backend='dogpile.cache.memory',
        )

        # replace existing backend since this may already be configured
        cache.CACHE_INVALIDATION_REGION.configure(
            backend='dogpile.cache.memory',
            expiration_time=None,
            replace_existing_backend=True,
        )

        self.region_name = uuid.uuid4().hex
        self.region0 = cache.create_region('test_region')
        self.region1 = cache.create_region('test_region')
        cache.configure_cache(region=self.region0)
        cache.configure_cache(region=self.region1)

        # TODO(dstanek): this should be a mock entrypoint
        self.cache_dict = {}
        self.backend = memory.MemoryBackend({'cache_dict': self.cache_dict})
        self.region0.backend = self.backend
        self.region1.backend = self.backend

    def _assert_has_no_value(self, values):
        for value in values:
            self.assertIsInstance(value, dogpile.NoValue)

    def test_singular_methods_when_invalidating_the_region(self):
        key = uuid.uuid4().hex
        value = uuid.uuid4().hex

        # key does not exist
        self.assertIsInstance(self.region0.get(key), dogpile.NoValue)
        # make it exist
        self.region0.set(key, value)
        # ensure it exists
        self.assertEqual(value, self.region0.get(key))

        # invalidating region1 should invalidate region0
        self.region1.invalidate()
        self.assertIsInstance(self.region0.get(key), dogpile.NoValue)

    def test_region_singular_methods_delete(self):
        key = uuid.uuid4().hex
        value = uuid.uuid4().hex

        # key does not exist
        self.assertIsInstance(self.region0.get(key), dogpile.NoValue)
        # make it exist
        self.region0.set(key, value)
        # ensure it exists
        self.assertEqual(value, self.region0.get(key))
        # delete it
        self.region1.delete(key)
        # ensure it's gone
        self.assertIsInstance(self.region0.get(key), dogpile.NoValue)

    def test_multi_methods_when_invalidating_the_region(self):
        mapping = {uuid.uuid4().hex: uuid.uuid4().hex for _ in range(4)}
        keys = list(mapping.keys())
        values = [mapping[k] for k in keys]

        # keys do not exist
        self._assert_has_no_value(self.region0.get_multi(keys))
        # make them exist
        self.region0.set_multi(mapping)
        # ensure they exist
        self.assertEqual(values, self.region0.get_multi(keys))
        # check using the singular get method for completeness
        self.assertEqual(mapping[keys[0]], self.region0.get(keys[0]))

        # invalidating region1 should invalidate region0
        self.region1.invalidate()
        # ensure they are gone
        self._assert_has_no_value(self.region0.get_multi(keys))

    def test_region_multi_methods_delete(self):
        mapping = {uuid.uuid4().hex: uuid.uuid4().hex for _ in range(4)}
        keys = list(mapping.keys())
        values = [mapping[k] for k in keys]

        # keys do not exist
        self._assert_has_no_value(self.region0.get_multi(keys))
        # make them exist
        self.region0.set_multi(mapping)
        # ensure they exist
        keys = list(mapping.keys())
        self.assertEqual(values, self.region0.get_multi(keys))
        # check using the singular get method for completeness
        self.assertEqual(mapping[keys[0]], self.region0.get(keys[0]))

        # delete them
        self.region1.delete_multi(mapping.keys())
        # ensure they are gone
        self._assert_has_no_value(self.region0.get_multi(keys))

    def test_memoize_decorator_when_invalidating_the_region(self):
        memoize = cache.get_memoization_decorator('cache', region=self.region0)

        @memoize
        def func(value):
            return value + uuid.uuid4().hex

        key = uuid.uuid4().hex

        # test get/set
        return_value = func(key)
        # the values should be the same since it comes from the cache
        self.assertEqual(return_value, func(key))

        # invalidating region1 should invalidate region0
        self.region1.invalidate()
        new_value = func(key)
        self.assertNotEqual(return_value, new_value)

    def test_combination(self):
        memoize = cache.get_memoization_decorator('cache', region=self.region0)

        @memoize
        def func(value):
            return value + uuid.uuid4().hex

        key = uuid.uuid4().hex
        simple_value = uuid.uuid4().hex

        # test get/set using the decorator
        return_value = func(key)
        self.assertEqual(return_value, func(key))

        # test get/set using the singular methods
        self.region0.set(key, simple_value)
        self.assertEqual(simple_value, self.region0.get(key))

        # invalidating region1 should invalidate region0
        self.region1.invalidate()

        # ensure that the decorated function returns a new value
        new_value = func(key)
        self.assertNotEqual(return_value, new_value)

        # ensure that a get doesn't have a value
        self.assertIsInstance(self.region0.get(key), dogpile.NoValue)

    def test_direct_region_key_invalidation(self):
        """Invalidate by manually clearing the region key's value.

        NOTE(dstanek): I normally don't like tests that repeat application
        logic, but in this case we need to. There are too many ways that
        the tests above can erroneosly pass that we need this sanity check.
        """
        region_key = cache.RegionInvalidationManager(
            None, self.region0.name
        )._region_key
        key = uuid.uuid4().hex
        value = uuid.uuid4().hex

        # key does not exist
        self.assertIsInstance(self.region0.get(key), dogpile.NoValue)
        # make it exist
        self.region0.set(key, value)
        # ensure it exists
        self.assertEqual(value, self.region0.get(key))

        # test invalidation
        cache.CACHE_INVALIDATION_REGION.delete(region_key)
        self.assertIsInstance(self.region0.get(key), dogpile.NoValue)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fixtures

from keystone.identity.backends.ldap import common as common_ldap
from keystone.tests.unit import fakeldap


class LDAPDatabase(fixtures.Fixture):
    """A fixture for setting up and tearing down an LDAP database."""

    def __init__(self, dbclass=fakeldap.FakeLdap):
        self._dbclass = dbclass

    def setUp(self):
        super().setUp()
        self.clear()
        common_ldap.WRITABLE = True
        common_ldap._HANDLERS.clear()
        common_ldap.register_handler('fake://', self._dbclass)
        # TODO(dstanek): switch the flow here
        self.addCleanup(self.clear)
        self.addCleanup(common_ldap._HANDLERS.clear)
        self.addCleanup(self.disable_write)

    def disable_write(self):
        common_ldap.WRITABLE = False

    def clear(self):
        for shelf in fakeldap.FakeShelves:
            fakeldap.FakeShelves[shelf].clear()
